/* libspe - A wrapper library to adapt the JSRE SPU usage model to SPUFS
 * Copyright (C) 2005 IBM Corp.
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 *  License for more details.
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <elf.h>
#include <elf_loader.h>

#define __PRINTF(fmt, args...) { fprintf(stderr,fmt , ## args); }
#ifdef DEBUG
#define DEBUG_PRINTF(fmt, args...) __PRINTF(fmt , ## args)
#else
#define DEBUG_PRINTF(fmt, args...)
#endif

int
load_spe_elf (void *elf_start, void *ld_buffer, struct spe_ld_info *ld_info)
{
	Elf32_Ehdr *ehdr;
	static const unsigned char expected[EI_PAD] = {
		[EI_MAG0] = ELFMAG0,
		[EI_MAG1] = ELFMAG1,
		[EI_MAG2] = ELFMAG2,
		[EI_MAG3] = ELFMAG3,
		[EI_CLASS] = ELFCLASS32,
		[EI_DATA] = ELFDATA2MSB,
		[EI_VERSION] = EV_CURRENT,
		[EI_OSABI] = ELFOSABI_SYSV,
		[EI_ABIVERSION] = 0
	};
	Elf32_Phdr *phdr;
	Elf32_Phdr *ph;
	int num_load_seg = 0;

	DEBUG_PRINTF ("load_spe_elf(%p, %p)\n", elf_start, ld_buffer);
	ehdr = (Elf32_Ehdr *) elf_start;

	/* Validate ELF */
	if (memcmp (ehdr->e_ident, expected, EI_PAD) != 0)
	  {
		  DEBUG_PRINTF ("invalid ELF header.\n");
		  DEBUG_PRINTF ("expected 0x%016llX != 0x%016llX\n",
				*(long long *) expected, *(long long *) ehdr);
		  errno = EINVAL;
		  return -errno;
	  }

	/* Validate the machine type */
	if (ehdr->e_machine != 0x17)
	  {
		  DEBUG_PRINTF ("not an SPE ELF object");
		  errno = EINVAL;
		  return -errno;
	  }

	/* Validate ELF object type. */
	if (ehdr->e_type != ET_EXEC)
	  {
		  DEBUG_PRINTF ("invalid SPE ELF type.\n");
		  DEBUG_PRINTF ("SPU type %d != %d\n", ehdr->e_type, ET_EXEC);
		  errno = EINVAL;
		  DEBUG_PRINTF ("parse_spu_elf(): errno=%d.\n", errno);
		  return -errno;
	  }

	/* Start processing headers */
	phdr = (Elf32_Phdr *) ((char *) ehdr + ehdr->e_phoff);

	/*
	 * Load all PT_LOAD segments onto the SPU local store buffer.
	 */
	DEBUG_PRINTF ("Segments: 0x%x\n", ehdr->e_phnum);
	for (ph = phdr; ph < &phdr[ehdr->e_phnum]; ++ph)
	  {
		  switch (ph->p_type)
		    {
		    case PT_LOAD:
			    /* DEBUG_PRINTF ("PT_LOAD)\n"); */
			    /* Only LOAD non-zero segments. */
			    if (ph->p_filesz)
			      {
				      num_load_seg++;

				      DEBUG_PRINTF
					      ("SPE_LOAD %p (0x%x) -> %p (0x%x) (%i bytes)\n",
					       ld_buffer + ph->p_vaddr,
					       ph->p_vaddr,
					       elf_start + ph->p_offset,
					       ph->p_offset, ph->p_filesz);
				      memcpy (ld_buffer + ph->p_vaddr,
					      elf_start + ph->p_offset,
					      ph->p_filesz);
			      }
			    break;
		    case PT_NOTE:
			DEBUG_PRINTF("SPE_LOAD found PT_NOTE\n");
			break;
		    }
	  }
	if (num_load_seg == 0)
	  {
		  DEBUG_PRINTF ("no segments to load");
		  errno = EINVAL;
		  return -errno;
	  }

	/* Remember where the code wants to be started */
	ld_info->entry = ehdr->e_entry;
	DEBUG_PRINTF ("entry = 0x%x\n", ehdr->e_entry);

	return 0;

}
