/*
 * libspe - A wrapper library to adapt the JSRE SPU usage model to SPUFS
 * Copyright (C) 2005 IBM Corp.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 *  License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <sys/spe.h>
#include <spe.h>
#include <elf_loader.h>
#include <mfc.h>
#include <spe_exec.h>
#include <default_c99_handler.h>


#define __PRINTF(fmt, args...) { fprintf(stderr,fmt , ## args); }
#ifdef DEBUG
#define DEBUG_PRINTF(fmt, args...) __PRINTF(fmt , ## args)
#else
#define DEBUG_PRINTF(fmt, args...)
#endif


/*
 * Helpers
 *
 * */

struct thread_store
{
	pthread_t spe_thread;
	int fd_mem, fd_run, fd_mbox, fd_mbox_stat, fd_ibox, fd_ibox_stat, fd_wbox, fd_wbox_stat, fd_npc, fd_sig1, fd_sig2;
	int npc, ret_status;
	spe_gid_t group_id;
	void *mem_mmap_base;
};

struct group_store
{
	int mem_access , policy, priority, use_events;
};

/*Default SPE group options
 */
static struct group_store default_group = {SPE_MEM_SHARED, SPE_SCHED_OTHER, 0, 0 };

/*
 * int spe_ldr[]:
 * SPE code that performs the actual parameter setting:
 */
static int spe_ldr[] = {
	0x30fff203,
	0x30fff404,
	0x30fff605,
	0x30fff806,
	0x30fff000,
	0x35000000,
	0x00000000,
	0x00000000
};

/*
 * int spe_ldr_debug[]:
 * Debug variant of loader that executes a breakpoint instruction
 * before starting main SPE code:
 */
static int spe_ldr_debug[] = {
	0x30fff203,
	0x30fff404,
	0x30fff605,
	0x30fff806,
	0x30fff000,
	0x00003fff,
	0x35000000,
	0x00000000
};

/*
 * int dump_regs[]:
 * SPE code that saves register content to local store:
 */
static int dump_regs[] = {
        0x20a00000, 0x20a00201, 0x20a00402, 0x20a00603,
        0x20a00804, 0x20a00a05, 0x20a00c06, 0x20a00e07,
        0x20a01008, 0x20a01209, 0x20a0140a, 0x20a0160b,
        0x20a0180c, 0x20a01a0d, 0x20a01c0e, 0x20a01e0f,
        0x20a02010, 0x20a02211, 0x20a02412, 0x20a02613,
        0x20a02814, 0x20a02a15, 0x20a02c16, 0x20a02e17,
        0x20a03018, 0x20a03219, 0x20a0341a, 0x20a0361b,
        0x20a0381c, 0x20a03a1d, 0x20a03c1e, 0x20a03e1f,
        0x20a04020, 0x20a04221, 0x20a04422, 0x20a04623,
        0x20a04824, 0x20a04a25, 0x20a04c26, 0x20a04e27,
        0x20a05028, 0x20a05229, 0x20a0542a, 0x20a0562b,
        0x20a0582c, 0x20a05a2d, 0x20a05c2e, 0x20a05e2f,
        0x20a06030, 0x20a06231, 0x20a06432, 0x20a06633,
        0x20a06834, 0x20a06a35, 0x20a06c36, 0x20a06e37,
        0x20a07038, 0x20a07239, 0x20a0743a, 0x20a0763b,
        0x20a0783c, 0x20a07a3d, 0x20a07c3e, 0x20a07e3f,
        0x20a08040, 0x20a08241, 0x20a08442, 0x20a08643,
        0x20a08844, 0x20a08a45, 0x20a08c46, 0x20a08e47,
        0x20a09048, 0x20a09249, 0x20a0944a, 0x20a0964b,
        0x20a0984c, 0x20a09a4d, 0x20a09c4e, 0x20a09e4f,
        0x20a0a050, 0x20a0a251, 0x20a0a452, 0x20a0a653,
        0x20a0a854, 0x20a0aa55, 0x20a0ac56, 0x20a0ae57,
        0x20a0b058, 0x20a0b259, 0x20a0b45a, 0x20a0b65b,
        0x20a0b85c, 0x20a0ba5d, 0x20a0bc5e, 0x20a0be5f,
        0x20a0c060, 0x20a0c261, 0x20a0c462, 0x20a0c663,
        0x20a0c864, 0x20a0ca65, 0x20a0cc66, 0x20a0ce67,
        0x20a0d068, 0x20a0d269, 0x20a0d46a, 0x20a0d66b,
        0x20a0d86c, 0x20a0da6d, 0x20a0dc6e, 0x20a0de6f,
        0x20a0e070, 0x20a0e271, 0x20a0e472, 0x20a0e673,
        0x20a0e874, 0x20a0ea75, 0x20a0ec76, 0x20a0ee77,
        0x20a0f078, 0x20a0f279, 0x20a0f47a, 0x20a0f67b,
        0x20a0f87c, 0x20a0fa7d, 0x20a0fc7e, 0x20a0fe7f,
        0x00002000, 0x00000000, 0x00000000, 0x00000000
};

/*
 * int restore_regs[]:
 * SPE code that restores register content from local store:
 */
static int restore_regs[] = {
        0x30a00000, 0x30a00201, 0x30a00402, 0x30a00603,
        0x30a00804, 0x30a00a05, 0x30a00c06, 0x30a00e07,
        0x30a01008, 0x30a01209, 0x30a0140a, 0x30a0160b,
        0x30a0180c, 0x30a01a0d, 0x30a01c0e, 0x30a01e0f,
        0x30a02010, 0x30a02211, 0x30a02412, 0x30a02613,
        0x30a02814, 0x30a02a15, 0x30a02c16, 0x30a02e17,
        0x30a03018, 0x30a03219, 0x30a0341a, 0x30a0361b,
        0x30a0381c, 0x30a03a1d, 0x30a03c1e, 0x30a03e1f,
        0x30a04020, 0x30a04221, 0x30a04422, 0x30a04623,
        0x30a04824, 0x30a04a25, 0x30a04c26, 0x30a04e27,
        0x30a05028, 0x30a05229, 0x30a0542a, 0x30a0562b,
        0x30a0582c, 0x30a05a2d, 0x30a05c2e, 0x30a05e2f,
        0x30a06030, 0x30a06231, 0x30a06432, 0x30a06633,
        0x30a06834, 0x30a06a35, 0x30a06c36, 0x30a06e37,
        0x30a07038, 0x30a07239, 0x30a0743a, 0x30a0763b,
        0x30a0783c, 0x30a07a3d, 0x30a07c3e, 0x30a07e3f,
        0x30a08040, 0x30a08241, 0x30a08442, 0x30a08643,
        0x30a08844, 0x30a08a45, 0x30a08c46, 0x30a08e47,
        0x30a09048, 0x30a09249, 0x30a0944a, 0x30a0964b,
        0x30a0984c, 0x30a09a4d, 0x30a09c4e, 0x30a09e4f,
        0x30a0a050, 0x30a0a251, 0x30a0a452, 0x30a0a653,
        0x30a0a854, 0x30a0aa55, 0x30a0ac56, 0x30a0ae57,
        0x30a0b058, 0x30a0b259, 0x30a0b45a, 0x30a0b65b,
        0x30a0b85c, 0x30a0ba5d, 0x30a0bc5e, 0x30a0be5f,
        0x30a0c060, 0x30a0c261, 0x30a0c462, 0x30a0c663,
        0x30a0c864, 0x30a0ca65, 0x30a0cc66, 0x30a0ce67,
        0x30a0d068, 0x30a0d269, 0x30a0d46a, 0x30a0d66b,
        0x30a0d86c, 0x30a0da6d, 0x30a0dc6e, 0x30a0de6f,
        0x30a0e070, 0x30a0e271, 0x30a0e472, 0x30a0e673,
        0x30a0e874, 0x30a0ea75, 0x30a0ec76, 0x30a0ee77,
        0x30a0f078, 0x30a0f279, 0x30a0f47a, 0x30a0f67b,
        0x30a0f87c, 0x30a0fa7d, 0x30a0fc7e, 0x30a0fe7f,
        0x00002000, 0x00000000, 0x00000000, 0x00000000
};

static int
spe_open_files(struct thread_store *spe, const char *pathname)
{
	char filename[40];
#ifdef NOSYSCALL
	sprintf(filename, "%s/run", pathname);
	spe->fd_run = open(filename, O_RDONLY);
#endif
	sprintf(filename, "%s/mbox", pathname);
	spe->fd_mbox = open(filename, O_RDONLY);

	sprintf(filename, "%s/mbox_stat", pathname);
	spe->fd_mbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox", pathname);
	spe->fd_ibox = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox_stat", pathname);
	spe->fd_ibox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/wbox", pathname);
	spe->fd_wbox = open(filename, O_WRONLY);

	sprintf(filename, "%s/wbox_stat", pathname);
	spe->fd_wbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/npc", pathname);
	spe->fd_npc = open(filename, O_RDWR);
		
	sprintf(filename, "%s/signal1", pathname);
	spe->fd_sig1 = open(filename, O_WRONLY);
		
	sprintf(filename, "%s/signal2", pathname);
	spe->fd_sig2 = open(filename, O_WRONLY);
		
	if (spe->fd_mbox      < 0 || spe->fd_ibox      < 0 || spe->fd_wbox      < 0 ||
	    spe->fd_mbox_stat < 0 || spe->fd_ibox_stat < 0 || spe->fd_wbox_stat < 0 ||
	    spe->fd_npc       < 0 || spe->fd_sig1      < 0 || spe->fd_sig2      < 0) {
		DEBUG_PRINTF("Could not open all SPE files.\n");
		close(spe->fd_mbox);
		close(spe->fd_mbox_stat);
		close(spe->fd_ibox);
		close(spe->fd_ibox_stat);
		close(spe->fd_wbox);
		close(spe->fd_wbox_stat);
		close(spe->fd_npc);
		close(spe->fd_sig1);
		close(spe->fd_sig2);
		return -1;
	}
	return 0;
}


/**
 * Low-level API
 * 
 */
void *
spe_gid_setup(int mem_access, int policy, int priority, int use_events)
{
	struct group_store *group_store;

        group_store = calloc (1, sizeof *group_store);
	if (!group_store)
	{
		DEBUG_PRINTF ("Could not allocate group store\n");
		errno = ENOMEM;
	        return NULL;
	}
	
	group_store->mem_access = mem_access;
	group_store->policy = policy;
	group_store->priority = priority;
	group_store->use_events = use_events;
	
	return group_store;
}

void *
spe_setup (spe_gid_t gid, void *start, void *argp, void *envp, int flags)
{
	struct thread_store *thread_store;
	int rc, memfd;
	addr64 argp64, envp64;
	char memname[40], pathname[40];
	void *spe_ld_buf;
#ifdef SPE_NO_MMAP
	ssize_t count = 0, num = 0;
#endif
	struct spe_ld_info ld_info;
	struct spe_exec_params spe_params __attribute__ ((aligned (4096)));
	
	if(!gid)
	{
		gid=&default_group;
	}

	/* Allocate thread structure */

	thread_store = calloc (1, sizeof *thread_store);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("Could not allocate thread store\n");
		  errno = ENOMEM;
		  return NULL;
	  }

	argp64.ull = (unsigned long long) (unsigned long) argp;
	envp64.ull = (unsigned long long) (unsigned long) envp;

	thread_store->group_id = gid;
	
	/* Make the SPU Directory */

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	DEBUG_PRINTF ("create %s\n", pathname);

#ifdef NOSYSCALL
	rc = mkdir(pathname, S_IRUSR | S_IWUSR | S_IXUSR);
#else
	rc = spe_create(pathname, 0, S_IRUSR | S_IWUSR | S_IXUSR);
	thread_store->fd_run = rc;
#endif
	if (rc < 0)
	  {
		  DEBUG_PRINTF ("Could not create SPE %s\n", pathname);
		  return NULL;
	  }

	sprintf (memname, "%s/mem", pathname);

	/* Check SPE */

	memfd = open (memname, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	if (memfd < 0)
	  {
		  DEBUG_PRINTF ("Could not open SPE mem file.\n");
		  return NULL;
	  }

	/* Prepare Loader */
#ifdef SPE_NO_MMAP	
	spe_ld_buf = malloc (LS_SIZE);
#else
	spe_ld_buf = mmap(0, LS_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0);
	thread_store->mem_mmap_base = spe_ld_buf;
	thread_store->fd_mem = memfd;
#endif
	if (!spe_ld_buf)
	  {
		  DEBUG_PRINTF ("Could not allocate SPE memory. \n");
		  errno = ENOMEM;
		  return NULL;
	  }

	memset(spe_ld_buf, 0, LS_SIZE);

	rc = load_spe_elf (start, spe_ld_buf, &ld_info);
	if (rc != 0)
	  {
		  DEBUG_PRINTF ("Load SPE ELF failed..\n");
		  return NULL;
	  }

	/* Add SPE exec program */

	if (!(flags & SPE_PTRACE_ME))
	  {
		  DEBUG_PRINTF ("Add exec prog dst:0x%04x size:0x%04x\n",
			        SPE_LDR_START, sizeof (spe_ldr));
		  memcpy (spe_ld_buf + SPE_LDR_START, &spe_ldr,
			  sizeof (spe_ldr));
	  }
	else
	  {
		  DEBUG_PRINTF ("Add debug exec prog dst:0x%04x size:0x%04x\n",
			        SPE_LDR_START, sizeof (spe_ldr_debug));
		  memcpy (spe_ld_buf + SPE_LDR_START, &spe_ldr_debug,
			  sizeof (spe_ldr_debug));
	  }

	/* Add SPE exec parameters */

	spe_params.gpr0[0] = ld_info.entry;
	spe_params.gpr0[1] = 0;
	spe_params.gpr0[2] = 0;
	spe_params.gpr0[3] = 0;
	
	spe_params.gpr3[0] = 0;
	spe_params.gpr3[1] = 0;
	spe_params.gpr3[2] = 0;
	spe_params.gpr3[3] = 0;
	
	spe_params.gpr4[0] = argp64.ui[0];
	spe_params.gpr4[1] = argp64.ui[1];
	spe_params.gpr4[2] = 0;
	spe_params.gpr4[3] = 0;
	
	spe_params.gpr5[0] = envp64.ui[0];
	spe_params.gpr5[1] = envp64.ui[1];
	spe_params.gpr5[2] = 0;
	spe_params.gpr5[3] = 0;

	spe_params.gpr6[0] = 0;
	spe_params.gpr6[1] = 0;
	spe_params.gpr6[2] = 0;
	spe_params.gpr6[3] = 0;

	DEBUG_PRINTF ("Add exec param dst:0x%04x size:0x%04x\n",
		      SPE_PARAM_START, sizeof (spe_params));
	memcpy (spe_ld_buf + SPE_PARAM_START, &spe_params,
		sizeof (spe_params));

	/* Syscall data element in PPE mem for newlib */
	memcpy (spe_ld_buf + SPE_SYSCALL_ADDR, &envp64.ull,
			                sizeof (envp64.ull));
#ifdef SPE_NO_MMAP
	/* Copy SPE image to SPUfs */
	do
	  {
		  num = write (memfd, spe_ld_buf + count, LS_SIZE - count);
		  if (num == -1)
		    {
			    DEBUG_PRINTF ("Transfer SPE ELF failed..\n");
			    return NULL;
		    }

		  count += num;
	  }
	while (count < LS_SIZE && num);
	close (memfd);

	/* Free the SPE Buffer */
	free (spe_ld_buf);
#endif

	rc = spe_open_files(thread_store, pathname);
	if (rc) {
		perror("open files\n");
		return NULL;
	}

	sync();

	return thread_store;
}

void
spe_cleanup (void *ptr)
{
	struct thread_store *thread_store = ptr;
	char pathname[40];

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	close (thread_store->fd_run);
	close (thread_store->fd_mbox);
	close (thread_store->fd_mbox_stat);
	close (thread_store->fd_ibox);
	close (thread_store->fd_ibox_stat);
	close (thread_store->fd_wbox);
	close (thread_store->fd_wbox_stat);
	close (thread_store->fd_npc);
	close (thread_store->fd_sig1);
	close (thread_store->fd_sig2);

#ifndef SPE_NO_MMAP
	munmap(thread_store->mem_mmap_base,LS_SIZE);
	close (thread_store->fd_mem);
#endif

#if defined(NOSYSCALL) && !defined(NODIRREMOVE)
	rmdir (pathname);
#endif

	free (thread_store);
}

unsigned int
set_npc (void *ptr, unsigned int npc)
{
	struct thread_store *thread_store = ptr;
	int ret;
	char npcText[9];

	sprintf(npcText,"0x%06x\n",npc);
	ret = write(thread_store->fd_npc, npcText, 9);
	if (ret < 0)
        {
                return -EINVAL;
        }
	return npc;
}

#ifdef NOSYSCALL
int
do_spe_run (void *ptr)
{
	struct thread_store *thread_store = ptr;
	int runfd;

	struct spufs_run_arg
	{
		unsigned int npc;	/* inout: Next Program Counter */
		unsigned short code;	/* out:   SPU status */
		unsigned short status;
	};
	struct spufs_run_arg arg = { SPE_LDR_PROG_start, 0, 0 };
	int ret;

        DEBUG_PRINTF ("do_spe_run\n");
	runfd=thread_store->fd_run;

	ret = set_npc(thread_store, arg.npc);
	if (ret < 0)
	{
		DEBUG_PRINTF ("Could not set NPC in NPC file.\n");
		return -EINVAL;
	}

	do
        {
#if 0
		ret = ioctl (runfd, _IOWR ('s', 0, struct spufs_run_arg),
                                &arg) & 0xff;
#else
		ret = read(runfd, &arg, 8);
#endif
                DEBUG_PRINTF ("SPE (0x%06x) (0x%04x) (0x%04x).\n"
                                ,arg.npc, arg.code,arg.status);

                if ( ret < 0)
                {
                          DEBUG_PRINTF ("Could not ioctl() on SPE run file.\n");
			  return -EINVAL;
                }
		else if ((arg.code&0xff00)==0x2100) {
			/* Per SPU-ABI (RFC 22-4), stop code
			 * 0x2100 is set aside for SPU-C99  
			 * emulation.
			 */
			DEBUG_PRINTF ("C99 Call data: 0x%06x.\n" ,arg.npc);
			default_c99_handler(thread_store->mem_mmap_base, arg.npc); 
		}

        } while ((arg.code & 0xff00) == 0x2100 || arg.code == 0x3fff);

	DEBUG_PRINTF ("SPE thread result: %08x:%04x:%04x\n", arg.npc,
		      arg.code, arg.status);
	
	/* Save status & npc */	
	thread_store->npc = arg.npc;
	thread_store->ret_status = arg.code;

	return arg.code;
}
#else
int
do_spe_run (void *ptr)
{
	struct thread_store *thread_store = ptr;
	int runfd;

	unsigned int npc;
	unsigned int code;
	unsigned int status;

	npc = SPE_LDR_PROG_start;

	int ret;

        DEBUG_PRINTF ("do_spe_run\n");
	runfd = thread_store->fd_run;

	do {
		ret = spe_run(runfd, &npc, &status);
		code = ret >> 16;
                DEBUG_PRINTF ("SPE (0x%06x) (0x%04x) (0x%04x).\n",
			      		npc, code, ret & 0xffff);

                if (ret < 0) {
                          DEBUG_PRINTF ("Could not run on SPE: %s.\n",
				strerror(errno));
			  return -errno;
                } else if ((code&0xff00)==0x2100) {
			/* Per SPU-ABI (RFC 22-4), stop code
                         * 0x2100 is set aside for SPU-C99
                         * emulation.
                         */
			DEBUG_PRINTF ("C99 Call data: 0x%06x.\n", npc);
			default_c99_handler(thread_store->mem_mmap_base, npc); 
		}

        } while ((code & 0xff00) == 0x2100 || code == 0x3fff);

	DEBUG_PRINTF ("SPE thread result: %08x:%04x\n", npc, code);
	
	/* Save status & npc */	
	thread_store->npc = npc;
	thread_store->ret_status = code;

	return code;
}
#endif

void *load_binary(char *bin)
{
        int binfd;
        struct stat statbuf;
        int ret;
        void *pos, *buf;
        off_t size;

        binfd = open(bin, O_RDONLY);
        if (binfd < 0)
                return NULL;

        ret = fstat(binfd, &statbuf);
        if (ret < 0)
                return NULL;

	buf=malloc(statbuf.st_size);

        if (!buf)
                return NULL;
	
        pos = buf;
        size = statbuf.st_size;

        do {
                ret = read(binfd, pos, size);
                if (ret > 0) {
                        pos += ret;
                        size -= ret;
                }
                if (ret < 0)
                        return NULL;
        } while (size > 0 && ret);

        return buf;
}


/**
 * Library API
 * 
 */

spe_gid_t
spe_create_group (int mem_access, int policy, int priority, int use_events)
{
	struct group_store *group_store;

	DEBUG_PRINTF ("spu_create_group(0x%x, 0x%x, 0x%x, 0x%x)\n",
			                  mem_access, policy, priority, use_events);
	
	group_store=spe_gid_setup(mem_access, policy, priority, use_events);
	DEBUG_PRINTF("gid is %p\n", group_store);

	return group_store;
}

int
spe_group_max (spe_gid_t spe_gid)
{
	DEBUG_PRINTF ("spu_group_max(0x%x)\n", spe_gid);

	return 8;
}


static void *
spe_thread (void *ptr)
{
	return (void *) (unsigned long) do_spe_run (ptr);
}

speid_t
spe_create_thread (spe_gid_t gid, void *start, void *argp, void *envp, int mask,
		   int flags)
{
	struct thread_store *thread_store;
	int rc;

	DEBUG_PRINTF ("spe_create_thread(%p, %p, %p, %p, 0x%x, 0x%x)\n",
		      gid, start, argp, envp, mask, flags);

	thread_store = spe_setup (gid, start, argp, envp, flags);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("spe_setup failed\n");
		  errno=EINVAL;
		  return NULL;
	  }
	
	rc = pthread_create (&thread_store->spe_thread, NULL,
			     spe_thread, thread_store);
	if (rc)
	  {
		  perror ("pthread_create");
		  return NULL;
	  }

	return thread_store;
}

int
spe_wait (speid_t speid, int *status, int options)
{
	struct thread_store *thread_store = speid;
	void *pthread_rc;
	int rc;

	DEBUG_PRINTF ("spu_wait(0x%x, %p, 0x%x)\n", speid, status, options);

	rc = pthread_join (thread_store->spe_thread, &pthread_rc);
	*status = (int) (unsigned long) pthread_rc;

	spe_cleanup (thread_store);

	DEBUG_PRINTF ("Thread ended.\n");
	return rc;
}

int
spe_wait_save (speid_t speid, struct spe_ctx *spe_context)
{
	struct thread_store *thread_store = speid;
	void *pthread_rc;
	int rc, ret;

	struct spufs_run_arg
	{
		unsigned int npc;       /* inout: Next Program Counter */
	        unsigned short code;    /* out:   SPU status */
	        unsigned short status;
        };
        struct spufs_run_arg arg = { 0, 0, 0 };

	rc = pthread_join (thread_store->spe_thread, &pthread_rc);
	
#ifdef SPE_NO_MMAP
        spe_cleanup (thread_store);
	
	DEBUG_PRINTF ("spe_wait_save(): This function is only available on mmapped spes.\n");
	
	return -1;
#else
	if (!spe_context->ls)
	{
		DEBUG_PRINTF ("spe_wait_save(): No LS to copy to.\n");
		return -1;
	}

	/* Save local store */
	memcpy (spe_context->ls, thread_store->mem_mmap_base, LS_SIZE);

	/* Save results */
	spe_context->npc = thread_store->npc;
	spe_context->status = thread_store->ret_status;
	
	/* Upload dump register code */
	memcpy (thread_store->mem_mmap_base + 0x1000, &dump_regs, sizeof(dump_regs));
	
	/* Run the register dumper */
	set_npc(thread_store, 0x1000);
	ret = read(thread_store->fd_run, &arg, 8);

	/* Save the registers */
	memcpy (&spe_context->reg, thread_store->mem_mmap_base + 0x10000, sizeof(spe_context->reg));

#endif

	spe_cleanup (thread_store);

	DEBUG_PRINTF ("Thread ended.\n");
	return rc;
}


int
spe_kill (speid_t speid, int sig)
{
	struct thread_store *thread_store = speid;
	int rc;

	thread_store=speid;	

	rc = pthread_kill (thread_store->spe_thread, sig);

	return rc;
}

int
spe_mbox_read(speid_t speid, unsigned int *val)
{
	struct thread_store *thread_store = speid;
	int rc;

	rc = read(thread_store->fd_mbox, val, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

int
spe_ibox_read(speid_t speid, unsigned int *val)
{
	struct thread_store *thread_store = speid;
	int rc;

	rc = read(thread_store->fd_ibox, val, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

int
spe_wbox_write(speid_t speid, unsigned int val)
{
	struct thread_store *thread_store = speid;
	int rc;

	rc = write(thread_store->fd_wbox, &val, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

void
*spe_get_ls (speid_t speid)
{
	struct thread_store *thread_store = speid;
	
#ifdef SPE_NO_MMAP
	errno = EFAULT;
	return -1;
#else
	return thread_store->mem_mmap_base;
#endif
}

void
*spe_get_ps (speid_t speid)
{
	return speid;
}

/*
 * mfc.h direct call-ins
 *
 * */

int _put_mfc_outbox (void *spu_ps_addr ,unsigned int data)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc;
	
	rc = write(thread_store->fd_wbox, &data, 4);
	if (rc == 4)
		rc = 0;

	return rc;

}

int _stat_mfc_outbox (void *spu_ps_addr)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc, ret;

	rc = read(thread_store->fd_wbox_stat, &ret, 1);
	if (rc != 4)
		ret = -1;

	return ret;

}

int _get_mfc_inbox (void *spu_ps_addr)
{
        struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_mbox, &ret, 1);
        if (rc != 4)
	        ret = -1;

        return ret;

}


int _stat_mfc_inbox (void *spu_ps_addr)
{
	struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_mbox_stat, &ret, 1);
        if (rc != 4)
                ret = -1;

        return ret;

}


int _stat_mfc_inbox_intr (void *spu_ps_addr)
{
        struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_ibox_stat, &ret, 1);
        if (rc != 4)
                ret = -1;

        return ret;

}


int _send_mfc_sig (void *spu_ps_addr, unsigned int signal_reg, unsigned int data)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc;

	rc = write(thread_store->fd_sig1, &data, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

